package com.jotuntech.sketcher.client.voice.aec;

public class AEC {
	/* dB Values */
	final static float M0dB = 1.0f;
	final static float M3dB = 0.71f;
	final static float M6dB = 0.50f;
	final static float M9dB = 0.35f;
	final static float M12dB = 0.25f;
	final static float M18dB = 0.125f;
	final static float M24dB = 0.063f;

	/* dB values for 16bit PCM */
	/* MxdB_PCM = 32767 * 10 ^(x / 20) */
	final static float M10dB_PCM = 10362.0f;
	final static float M20dB_PCM = 3277.0f;
	final static float M25dB_PCM = 1843.0f;
	final static float M30dB_PCM = 1026.0f;
	final static float M35dB_PCM = 583.0f;
	final static float M40dB_PCM = 328.0f;
	final static float M45dB_PCM = 184.0f;
	final static float M50dB_PCM = 104.0f;
	final static float M55dB_PCM = 58.0f;
	final static float M60dB_PCM = 33.0f;
	final static float M65dB_PCM = 18.0f;
	final static float M70dB_PCM = 10.0f;
	final static float M75dB_PCM = 6.0f;
	final static float M80dB_PCM = 3.0f;
	final static float M85dB_PCM = 2.0f;
	final static float M90dB_PCM = 1.0f;

	final static float MAXPCM = 32767.0f;

	/* Design final staticants (Change to fine tune the algorithms */

	/* The following values are for hardware AEC and studio quality
	 * microphone */

	/* maximum NLMS filter length in taps. A longer filter length gives
	 * better Echo Cancellation, but slower convergence speed and
	 * needs more CPU power (Order of NLMS is linear) */
	final static int NLMS_LEN = 80 * 8;

	/* convergence speed. Range: >0 to <1 (0.2 to 0.7). Larger values give
	 * more AEC in lower frequencies, but less AEC in higher frequencies. */
	final static float Stepsize = 0.7f;

	/* minimum energy in xf. Range: M70dB_PCM to M50dB_PCM. Should be equal
	 * to microphone ambient Noise level */
	final static float Min_xf = M75dB_PCM;

	/* Double Talk Detector Speaker/Microphone Threshold. Range <=1
	 * Large value (M0dB) is good for Single-Talk Echo cancellation,
	 * small value (M12dB) is good for Doulbe-Talk AEC */
	final static float GeigelThreshold = M6dB;

	/* Double Talk Detector hangover in taps. Not relevant for Single-Talk
	 * AEC */
	final static int Thold = 30 * 8;

	/* for Non Linear Processor. Range >0 to 1. Large value (M0dB) is good
	 * for Double-Talk, small value (M12dB) is good for Single-Talk */
	final static float NLPAttenuation = M12dB;

	// Extention in taps to reduce mem copies
	final static int NLMS_EXT = 10 * 8;

	// block size in taps to optimize DTD calculation
	final static int DTD_LEN = 16;

	// Time domain Filters
	IIRHP hp00, hp1;             // DC-level remove Highpass)
	FIRHP13 hp0;                 // 300Hz cut-off Highpass
	IIR1 Fx, Fe;                  // pre-whitening Highpass for x, e

	// Geigel DTD (Double Talk Detector)
	float max_max_x;               // max(|x[0]|, .. |x[L-1]|)
	int hangover;
	// optimize: less calculations for max()
	float max_x[];
	int dtdCnt;
	int dtdNdx;

	// NLMS-pw
	float x[];  // tap delayed loudspeaker signal
	float xf[]; // pre-whitening tap delayed signal
	float w[];             // tap weights
	int j;                        // optimize: less memory copies
	int lastupdate;               // optimize: iterative dotp(x,x)
	double dotp_xf_xf;            // double to avoid loss of precision
	double Min_dotp_xf_xf;
	float s0avg;

	/* Vector Dot Product */
	private float dotp(float a[], float b[], int boffset) {
		float sum0 = 0.0f, sum1 = 0.0f;
		int j;

		for (j = 0; j < NLMS_LEN; j += 2) {
			// optimize: partial loop unrolling
			sum0 += a[j] * b[j + boffset];
			sum1 += a[j + 1] * b[j + 1 + boffset];
		}
		return sum0 + sum1;
	}

	public float getAmbient() {
		return s0avg;
	};

	public void setAmbient(float Min_xf) {
		dotp_xf_xf = Min_dotp_xf_xf = NLMS_LEN * Min_xf * Min_xf;
	};

	public AEC() {
		hp00 = new IIRHP();
		hp1 = new IIRHP();
		hp0 = new FIRHP13();
		Fx = new IIR1();
		Fe = new IIR1();

		max_x = new float[NLMS_LEN / DTD_LEN];

		x = new float[NLMS_LEN + NLMS_EXT];
		xf = new float[NLMS_LEN + NLMS_EXT];
		w = new float[NLMS_LEN];

		max_max_x = 0.0f;
		hangover = 0;
		dtdCnt = dtdNdx = 0;

		j = NLMS_EXT;
		lastupdate = 0;
		s0avg = M80dB_PCM;
		setAmbient(Min_xf);
	}

	public float nlms_pw(float mic, float spk, boolean update) {
		float d = mic;      	        // desired signal
		x[j] = spk;
		xf[j] = Fx.highpass(spk);     // pre-whitening of x

		// calculate error value
		// (mic signal - estimated mic signal from spk signal)
		float e = d - dotp(w, x, j);
		float ef = Fe.highpass(e);    // pre-whitening of e
		// optimize: iterative dotp(xf, xf)
		dotp_xf_xf += (xf[j] * xf[j] - xf[j + NLMS_LEN - 1] * xf[j + NLMS_LEN - 1]);
		if (update) {
			// calculate variable step size
			float mikro_ef = (float) (Stepsize * ef / dotp_xf_xf);

			// update tap weights (filter learning)
			int i;
			for (i = 0; i < NLMS_LEN; i += 2) {
				// optimize: partial loop unrolling
				w[i] += mikro_ef * xf[i + j];
				w[i+1] += mikro_ef * xf[i + j + 1];
			}
		}

		if (--j < 0) {
			// optimize: decrease number of memory copies
			j = NLMS_EXT;
			System.arraycopy(x, 0, x, j + 1, NLMS_LEN - 1);
			System.arraycopy(xf, 0, xf, j + 1, NLMS_LEN - 1);
		}

		return e;
	}

	public boolean dtd(float d, float x) {
		// optimized implementation of max(|x[0]|, |x[1]|, .., |x[L-1]|):
		// calculate max of block (DTD_LEN values)
		x = Math.abs(x);
		if (x > max_x[dtdNdx]) {
			max_x[dtdNdx] = x;
			if (x > max_max_x) {
			max_max_x = x;
			}
		}
		if (++dtdCnt >= DTD_LEN) {
			dtdCnt = 0;
			// calculate max of max
			max_max_x = 0.0f;
			for (int i = 0; i < NLMS_LEN/DTD_LEN; ++i) {
			if (max_x[i] > max_max_x) {
				max_max_x = max_x[i];
			}
			}
			// rotate Ndx
			if (++dtdNdx >= NLMS_LEN/DTD_LEN) dtdNdx = 0;
			max_x[dtdNdx] = 0.0f;
		}

		// The Geigel DTD algorithm with Hangover timer Thold
		if (Math.abs(d) >= GeigelThreshold * max_max_x) {
			hangover = Thold;
		}

		if(hangover != 0) --hangover;

		return (hangover > 0);
	}


	public float doAEC(int d, int x) {
		float s0 = (float)d;
		float s1 = (float)x;

		// Mic Highpass Filter - to remove DC
		s0 = hp00.highpass(s0);

		// Mic Highpass Filter - telephone users are used to 300Hz cut-off
		s0 = hp0.highpass(s0);

		// ambient mic level estimation
		s0avg += 1e-4f * (Math.abs(s0) - s0avg);

		// Spk Highpass Filter - to remove DC
		s1 = hp1.highpass(s1);

		// Double Talk Detector
		boolean update = !dtd(s0, s1);

		// Acoustic Echo Cancellation
		s0 = nlms_pw(s0, s1, update);

		// Acoustic Echo Suppression
		if (update) {
			// Non Linear Processor (NLP): attenuate low volumes
			s0 *= NLPAttenuation;
		}

		return s0;
	}
}
