package com.jotuntech.sketcher.client.voice.aec;

/* 13 taps FIR Finite Impulse Response filter
 * Coefficients calculated with
 * www.dsptutor.freeuk.com/KaiserFilterDesign/KaiserFilterDesign.html
 */

public class FIRHP13 {
	private final static float a[] = {
		// Kaiser Window FIR Filter, Filter type: High pass
		// Passband: 300.0 - 4000.0 Hz, Order: 12
		// Transition band: 100.0 Hz, Stopband attenuation: 10.0 dB
		-0.043183226f, -0.046636667f, -0.049576525f, -0.051936015f,
		-0.053661242f, -0.054712527f, 0.82598513f, -0.054712527f,
		-0.053661242f, -0.051936015f, -0.049576525f, -0.046636667f,
		-0.043183226f, 0.0f
	};

	private float z[];

	public FIRHP13() {
		z = new float[14];
	}

	float highpass(float in) {
		System.arraycopy(z, 0, z, 1, z.length - 1);
		for(int i = a.length - 2; i >= 0; i--) {
			z[i + 1] = z[i];
		}
		z[0] = in;

		float sum0 = 0.0f, sum1 = 0.0f;
		int j;
		for (j = 0; j < a.length; j += 2) {
			// optimize: partial loop unrolling
			sum0 += a[j] * z[j];
			sum1 += a[j + 1] * z[j + 1];
		}

		return sum0 + sum1;
	}
}
