package com.jotuntech.sketcher.client.voice.aec;

/* Recursive two pole IIR Infinite Impulse Response filter
 * Coefficients calculated with
 * http://www.dsptutor.freeuk.com/IIRFilterDesign/IIRFiltDes102.html
 */

public class IIR2 {
	private final static float a[] = { 0.29289323f, -0.58578646f, 0.29289323f };
	private final static float b[] = { 1.3007072E-16f, 0.17157288f };

	private float x[], y[];

	public IIR2() {
		x = new float[2];
		y = new float[2];
	}

	float highpass(float in) {
		// Butterworth IIR filter, Filter type: HP
		// Passband: 2000 - 4000.0 Hz, Order: 2
		float out =
			a[0] * in +
			a[1] * x[0] +
			a[2] * x[1] -
			b[0] * y[0] -
			b[1] * y[1];

		x[1] = x[0];
		x[0] = in;
		y[1] = y[0];
		y[0] = out;

		return out;
	}
}